# -*- perl -*-

#
# $Id: UTM.pm,v 1.3 2002/05/12 23:43:32 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 2002 Slaven Rezic. All rights reserved.
#
#      This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
# Original formulas from:
#
#  gpsman --- GPS Manager: a manager for GPS receiver data
#
#  Copyright (c) 2001 Miguel Filgueiras (mig@ncc.up.pt) / Universidade do Porto
#
# (file posncomp.tcl)
#

package Karte::UTM;

use strict;
use vars qw($UTMlat0 $UTMk0 $UPSk0);
use POSIX qw(floor pow atan tan);

$UTMlat0 = 0;
$UTMk0 = 0.9996;
$UPSk0 = 0.994;

=item DegreesToUTM

Convert position in lat/long signed degrees and given datum to UTM/UPS.

=cut

sub DegreesToUTM {
    my($lat, $long, $datum) = @_;
    my($zn, $ze, $x, $y);
    if ($lat >= -80 && $lat <= 84) {
        $zn = 67 + int(floor(($lat+80)/8.0));
        # skip I, O
        $zn++ if ($zn > 72);
        $zn++ if ($zn > 78);
        $zn = chr($zn);
        my $long0 =  6*int(floor($long/6.0))+3;
        $ze = sprintf "%02d", int(floor(($long0+183)/6.0));
        my(@cs) = ConvToTM($lat, $long, $UTMlat0, $long0, $UTMk0, $datum);#XXXsub
        $x = sprintf "%.0f", 5e5+$cs[0];
	$y = $cs[1];
        if ($lat < 0) { $y = int(1e7+$y) }
    } else {
        $ze = "00";
        if ($lat > 0) {
            if ($long < 0) { $zn = "Y" } else { $zn = "Z" }
        } else {
            if ($long < 0) { $zn = "A" } else { $zn = "B" }
        }
        ($x, $y) = ConvToUPS($lat, $long, $datum);
    }
    ($ze, $zn, $x, $y);
}

=item ConvToTM

Convert to TM.
$lat0, $long0: centre coordinates in signed degrees.
$k0: scale at central meridian.

=cut

sub ConvToTM {
    my($lat, $long, $lat0, $long0, $k0, $datum) = @_;

    my($a, $f) = EllipsdData($datum);
    my $es = $f*(2-$f);
    my $phi = $lat*0.01745329251994329576;
    my $lambda = $long*0.01745329251994329576;
    my $phi0 = $lat0*0.01745329251994329576;
    my $lambda0 = $long0*0.01745329251994329576;
    my $m0 = TMAux($phi0, $a, $es);
    my $m  = TMAux($phi, $a, $es);
    my $et2 = $es/(1-$es);
    my $n = sin($phi); $n = $a/sqrt(1-$es*$n*$n);
    my $t = tan($phi); $t = $t*$t;
    my $c = cos($phi);
    my $A = ($lambda-$lambda0)*$c; $c = $et2*$c*$c;
    my $x = sprintf "%.0f", ($k0*$n*($A+(1-$t+$c)*$A*$A*$A/6+
				     (5-18*$t+$t*$t+72*$c-58*$et2)*
				     $A*$A*$A*$A*$A/120));
    my $y = sprintf "%.0f", ($k0*($m-$m0+$n*tan($phi)*
				  ($A*$A/2+(5-$t+9*$c+4*$c*$c)*$A*$A*$A*$A/24+
				   (61-58*$t+$t*$t+600*$c-330*$et2)*
				   $A*$A*$A*$A*$A*$A/720)));
    ($x, $y);
}

=item ConvFromTM

Convert from TM.
$lat0, $long0: centre coordinates in signed degrees.
$k0: scale at central meridian.

=cut

sub ConvFromTM {
    my($x, $y, $lat0, $long0, $k0, $datum) = @_;

    my($a,$f) = EllipsdData($datum);
    my $es = $f*(2-$f);
    my $phi0 = $lat0*0.01745329251994329576;
    my $lambda0 = $long0*0.01745329251994329576;
    my $m0 = TMAux($phi0, $a, $es);
    my $m = $m0+1.0*$y/$k0;
    my $et2 = $es/(1-$es);
    my $e1 = sqrt(1-$es); $e1 = (1-$e1)/(1+$e1);
    my $mu = $m/($a*(1-$es/4-3*$es*$es/64-0.01953125*$es*$es*$es));
    my $phi1 = ($mu+(1.5*$e1-27*$e1*$e1*$e1/32)*sin($mu+$mu)+
		(1.3125-1.71875*$e1*$e1)*$e1*$e1*sin(4*$mu)+
		1.572916666667*$e1*$e1*$e1*sin(6*$mu)+
		+2.142578125*$e1*$e1*$e1*$e1*sin(8*$mu));
    my $c1 = cos($phi1); $c1 = $et2*$c1*$c1;
    my $t1 = tan($phi1); $t1 = $t1*$t1;
    my $n1 = sin($phi1);
    my $r1 = $a*(1-$es)/pow(1-$es*$n1*$n1, 1.5);
    $n1 = $a/sqrt(1-$es*$n1*$n1);
    my $d = $x/($n1*$k0);
    my $lat = ($phi1-$n1*tan($phi1)/$r1*
	       ($d*$d/2-(5+3*$t1+10*$c1-4*$c1*$c1-9*$et2)*
		$d*$d*$d*$d/24+
		(61+90*$t1+298*$c1+45*$t1*$t1-252*$et2-3*$c1*$c1)*
		$d*$d*$d*$d*$d*$d/720))*57.29577951308232087684;
    my $long = ($lambda0+($d-(1+($t1+$t1)+$c1)*$d*$d*$d/6+
			  (5-($c1+$c1)+28*$t1-3*$c1*$c1+8*$et2+24*$t1*$t1)*
			  $d*$d*$d*$d*$d/120)/cos($phi1))*57.29577951308232087684;
    ($lat, $long);
}

sub TMAux {
    my($phi, $a, $es) = @_;

    if (abs($phi) < 1e-20) { return 0 }
    my $es2 = $es*$es;
    my $es3 = $es2*$es;
    $a*((1-$es/4-0.046875*$es2-0.01953125*$es3)*$phi-
	(0.375*$es+0.09375*$es2+0.0439453125*$es3)*
	sin($phi+$phi)+
	(0.05859375*$es2+0.0439453125*$es3)*sin(4*$phi)-
	0.011393229166667*$es3*sin(6*$phi))
}

sub ConvToUPS {
    my($lat, $long, $datum) = @_;

    my($a, $f) = EllipsdData($datum);
    my $es = $f*(2-$f);
    my $lambda = $long*0.01745329251994329576;
    my $phi = abs($lat*0.01745329251994329576);
    my $e = sqrt($es);
    my $ee = $e*sin($phi);
    my $t = tan(0.78539816339744830961-$phi/2)/pow((1-$ee)/(1+$ee),($e/2));
    my $rho = 2*$a*$UPSk0*$t/sqrt(pow(1+$e,1+$e)*pow(1-$e,1-$e));
    my $x = sprintf "%.0f", ($rho*sin($lambda)+2e6);
    my $y = $rho*cos($lambda);

    if ($lat > 0) { $y = -$y }
    ($x, sprintf "%.0f", $y+2e6);
}

=item UTMToDegrees

Convert from UTM/UPS to lat/long in signed degrees.

=cut

sub UTMToDegrees {
    my($ze, $zn, $x, $y, $datum) = @_;

    if ($ze != 0) {
	my $long0 = -183+6.0*$ze;
        if ($zn le "M") {
            $y = $y-1e7;
        }
        $x = $x-5e5;
        ConvFromTM($x, $y, $UTMlat0, $long0, $UTMk0, $datum);
    } else {
        ConvFromUPS($zn, $x, $y, $datum);
    }
}

sub ConvFromUPS {
    my($zn, $x, $y, $datum) = @_;

    my($a, $f) = EllipsdData($datum);
    my $es = $f*(2-$f);
    my $e = sqrt($es);
    $x = $x-2e6;
    $y = $y-2e6;
    my $rho = sqrt($x*$x+$y*$y);
    my $t = $rho*sqrt(pow(1+$e,1+$e)*pow(1-$e,1-$e))/(2*$a*$UPSk0);
    my $lat = UPSAux($e, $t)*57.29577951308232087684;
    my $long;
    if (abs($y) > 1e-20) {
	$long = atan(abs(1.0*$x/$y))*57.29577951308232087684;
    } else {
        $long = 90;
        if ($x < 0) { $long = -$long }
    }
    if ($zn gt "M") {
        $y = -$y;
    }
    if ($y < 0) { $long = 180-$long }
    if ($x < 0) { $long = -$long }
    ($lat, $long);
}

sub UPSAux {
    my($e, $t) = @_;

    my $phi = 1.57079632679489661923-2*atan($t);
    my $old = 999.9e-99;
    my $maxIterations = 20;
    my $i = 0;
    while (abs(($phi-$old)/$phi) > 1e-8 && $i < $maxIterations) {
        $i++;
        $old = $phi;
        $phi = (1.57079632679489661923-
		2*atan($t*pow((1-$e*sin($phi))/((1+$e*sin($phi))),$e/2.0)));
    }
    $phi;
}

######################################################################

use vars qw(%GDATUM %ELLPSDDEF);

## datum definitions
# index of GDATUM is datum name
# each list contains:
#  - ellipsoid name
#  - dx, dy, dz
#  - reference id; if a number, it is the number used by GPStrans
#  - comment
#    (the following elements are not being used and may be absent)
#  - expected error estimate in metres: ex, ey, ez (-1 for unknown)
#  - number of satellite measurement stations (0 for unknown)
#  - min, max lat and min, max long in signed degrees (empty or "_" if unknown)

%GDATUM =
    (
     "WGS 84" => ["WGS 84", 0, 0, 0, 100, "", -1, -1, -1, 0],
    );

## ellipsoid definitions
# indexed by name
# each list has:
#  - a (semi-major axis in metre), invf (inverse of flattening)
#  - comment

## those after "WGS 66" until "Walbeck" were taken from PROJ4.0

%ELLPSDDEF =
    (
     "WGS 84" => [6378137.0, 298.257223563, ""],
    );


=item EllipsdData

Based on contribution by Stefan Heinen.
Yield the ellipsoid data for datum as a list with
    a - semi-major axis
    f - flattening
    dx, dy, dz - translation of center

=cut

sub EllipsdData {
    my $datum = shift || "WGS 84";

    my @d = @{ $GDATUM{$datum} }[0..3];
    my $def = $ELLPSDDEF{$d[0]};
    ($def->[0], 1.0/$def->[1], @d[1..$#d]);
}

return 1 if caller;

package main;

my($lat, $long) = @ARGV;
foreach my $ref (\$lat, \$long) {
    if ($$ref =~ /\s+/) {
	warn "DMS coord detected, converting to DDD\n";
	require Karte::Polar;
	$$ref = Karte::Polar::dms_string2ddd($$ref);
    }
}

warn join " ", Karte::UTM::DegreesToUTM($lat, $long), "\n";

__END__
