# -*- perl -*-

#
# $Id: Build.pm,v 1.7 2002/02/10 20:57:07 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 2001 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: slaven.rezic@berlin.de
# WWW:  http://www.rezic.de/eserte/
#

package Strassen::Build;
use strict;
use vars qw($VERSION);

$VERSION = sprintf("%d.%02d", q$Revision: 1.7 $ =~ /(\d+)\.(\d+)/);

package StrassenNetz;

use vars qw($C_NET_FILE_VERSION $C_NET_MAGIC);

$C_NET_FILE_VERSION = 1;
$C_NET_MAGIC = 'stnt';

require Storable;

# Format of mmap file:
# (all little endian longs)
#   at the top of the file: a version number
#
#   foreach coordinate
#     x of this coordinate
#     y of this coordinate
#     number of successors
#     foreach successor:
#       relative pointer to successor struct
#       distance from this point to successor point in m
#
# The coord2ptr file is a Storable file of a hash which does the
# coordinate string ("x,y") to relative pointer mapping.
#
sub create_mmap_net {
    my($self, $file_prefix, %args) = @_;
    $self->make_net_classic if (!$self->{Net});

    use constant SIZEOF_LONG => 4; # XXX only for intel

    my $coord2ptr          = {};
    my $coord2structlength = {};
    my $total_structlength = SIZEOF_LONG*2; # room for magic and version number

    my $hash_file = $self->filename_c_net_coord2ptr($file_prefix);
    my $mmap_file = $self->filename_c_net_mmap($file_prefix);
    my $net2name_file = $self->filename_c_net_net2name($file_prefix);

    # XXX these values are only for intel platforms valid!
    use constant LENGTH_HEADER => SIZEOF_LONG*3; # three longs for x, y, number of succ
    use constant LENGTH_SUCC   => SIZEOF_LONG*2; # two longs for pointer and distance

    # first pass: calculate structs and create $coord2ptr Hash
    while(my($coord, $val) = each %{ $self->{Net} }) {
	$coord2ptr->{$coord} = $total_structlength;
	my $this_structlength = LENGTH_HEADER + LENGTH_SUCC * scalar keys %$val;
	$total_structlength += $this_structlength;
    }

    # second pass: create mmap file
    open(F, "> $mmap_file") or die "Can't create $mmap_file: $!";
    binmode F;

    print F $C_NET_MAGIC;
    print F pack("V", $C_NET_FILE_VERSION);

    while(my($coord, $val) = each %{ $self->{Net} }) {
	my($x,$y) = split /,/, $coord;
	my $header = pack("VVV", $x, $y, scalar keys %$val);
	print F $header;
	while(my($succ, $dist) = each %{ $val }) {
	    my $ptr = $coord2ptr->{$succ};
	    if (!defined $ptr) {
		die "No pointer in coord2ptr hash for $succ";
	    }
	    my $succrecord = pack("VV", $ptr, $dist);
	    print F $succrecord;
	}
    }
    close F;

    if (!Storable::store($coord2ptr, $hash_file)) {
	die "Can't write coord2ptr hash to file $hash_file: $!";
    }

    if (!Storable::store($self->{Net2Name}, $net2name_file)) {
	die "Can't write net2name hash to file $net2name_file: $!";
    }

    1;
}

sub filename_c_net_coord2ptr {
    my($self, $file_prefix) = @_;
    $file_prefix . "_coord2ptr.st";
}

sub filename_c_net_mmap {
    my($self, $file_prefix) = @_;
    $file_prefix . "_net.mmap";
}

sub filename_c_net_net2name {
    my($self, $file_prefix) = @_;
    $file_prefix . "_net2name.mmap";
}

sub create_mmap_net_if_needed {
    my($self, $file_prefix, %args) = @_;
    my $doit = 0;
    if (   !-e $self->{Strassen}->{File}
	|| !-e $self->filename_c_net_coord2ptr($file_prefix)
	|| !-e $self->filename_c_net_mmap($file_prefix)
	|| !-e $self->filename_c_net_net2name($file_prefix)
       ) {
	$doit = 1;
    } elsif (   -M $self->{Strassen}->{File} < -M $self->filename_c_net_coord2ptr($file_prefix)
	     || -M $self->{Strassen}->{File} < -M $self->filename_c_net_mmap($file_prefix)
	     || -M $self->{Strassen}->{File} < -M $self->filename_c_net_net2name($file_prefix)
	    ) {
	$doit = 1;
    }

    if ($doit) {
	$self->create_mmap_net($file_prefix, %args);
    } else {
	1;
    }
}


1;

__END__
