# -*- perl -*-

#
# $Id: Kreuzungen.pm,v 1.1 2002/05/21 23:47:05 eserte Exp $
#
# Copyright (c) 1995-2001 Slaven Rezic. All rights reserved.
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, see the file COPYING.
#
# Author: Slaven Rezic (eserte@cs.tu-berlin.de)
#

package Strassen::Kreuzungen;

package Kreuzungen;
use strict;
#use AutoLoader 'AUTOLOAD';

# Argumente: entweder "Hash" oder "Strassen"
# "Hash" ist der Rckgabewert von Strassen::all_crossings (bei
# Verwendung von "hash" oder "hashpos" als RetType). Array wird noch nicht
# verwendet.
# Bei "Strassen" werden die Kreuzungen automatisch per all_crossings erstellt.
sub new {
    my($class, %args) = @_;
    my $self = {};
    my $rettype = 'hash';
    my $usecache = defined $args{UseCache} ? $args{UseCache} : 1;
    if (!exists $args{Hash}) {
	die "Missing arg for new Kreuzungen: either Hash or Strassen"
	    if !exists $args{Strassen};
	if ($args{WantPos}) {
	    $rettype .= "pos";
	    $self->{Strassen} = $args{Strassen};
	}
	my(%kreuzungen) = $args{Strassen}->all_crossings
	    (RetType => $rettype, UseCache => $usecache);
	$args{Hash} = \%kreuzungen;
    }
    $self->{Hash}  = $args{Hash};
    $self->{Array} = $args{Array};
    $self->{IsPos} = $args{WantPos};
    bless $self, $class;
}

# Nimmt an. dass alle Punkte aus der Strassen-Datei Kreuzungen (z.B. Orte)
# darstellen.
### AutoLoad Sub
sub new_from_strassen {
    my($class, %args) = @_;
    my $want_pos = $args{WantPos};
    my $self = {};
    $self->{Hash} = {};
    my $str = $args{Strassen};
    if (!$str) {
	die "Missing arg for new_from_strassen: Strassen";
    }
    $str->init;
    my $pos = 0;
    while(1) {
	my $ret = $str->next;
	last if !@{ $ret->[Strassen::COORDS] };
	foreach my $p (@{ $ret->[Strassen::COORDS] }) {
	    $self->{Hash}{$p} = ($want_pos ? $pos : $ret->[Strassen::NAME]);
	}
	$pos++;
    }
    if ($want_pos) {
	$self->{IsPos} = 1;
	$self->{Strassen} = $str;
    }
    bless $self, $class;
}

### Autoload Sub
sub add {
    my($self, %args) = @_;
    if (!exists $args{Hash}) {
	die "Missing argument Hash";
    }
    my $h = $self->{Hash};
    while(my($k,$v) = each %{ $args{Hash} }) {
	$h->{$k} = $v;
    }
}

### AutoLoad Sub
sub get {
    my($self, $coord) = @_;
    if ($self->{IsPos}) {
	[ map { $self->{Strassen}->get($_)->[Strassen::NAME] } @{ $self->{Hash}{$coord} } ];
    } else {
	$self->{Hash}{$coord}
    }
}

### AutoLoad Sub
sub make_grid {
    my($self, $grid_width, $grid_height) = @_;
    $self->{GridWidth}  = (defined $grid_width ? $grid_width : 1000);
    $self->{GridHeight} = (defined $grid_height ? $grid_height : $self->{GridWidth});
    $self->{Grid} = {};
    while(defined(my $k = each %{$self->{Hash}})) {
	my $grid = join(",", $self->grid(split(/,/, $k)));
	push @{$self->{Grid}{$grid}}, $k;
    }
}

### AutoLoad Sub
sub grid {
    my($self, $x, $y) = @_;
    my($gx,$gy) = (int($x/$self->{GridWidth}), int($y/$self->{GridHeight}));
    $gx-- if $x < 0;
    $gy-- if $y < 0;
    ($gx,$gy);
}

# Argumente:
#    $x, $y
#    IncludeDistance: Rckgabewert ist ein Feld mit den Elementen ["x,y", dist]
#    Grids (default 1): Anzahl der Grids, die vom Mittelpunkt abweichen drfen
# Ansonsten wird ein Feld mit den Elementen "x,y" zurckgegeben. In beiden
# Fllen ist das Feld nach der Entfernung von $x und $y sortiert (krzeste
# zuerst).
### AutoLoad Sub
sub nearest {
    my($self, $x, $y, %args) = @_;
    $self->make_grid if (!$self->{Grid});
    my $xy = join(",", $x, $y);
    my $grids = $args{Grids} || 1;
#   my $distance = $args{Distance};
# XXX distance wird noch ignoriert
#    if (!defined $distance) { $distance = 1000 }
    my($gridx, $gridy) = $self->grid($x, $y);
    my @res;
    for my $xx (-$grids .. $grids) {
	for my $yy (-$grids .. $grids) {
	    my $s = join(",", $gridx+$xx, $gridy+$yy);
	    if (defined $self->{Grid}{$s}) {
		push @res, @{$self->{Grid}{$s}};
	    }
	}
    }
    if ($args{IncludeDistance}) {
	foreach (@res) {
	    my $dist = Strassen::Util::strecke_s($_, $xy);
	    $_ = [$_, $dist];
	}
	sort { $a->[1] <=> $b->[1] } @res;
    } else {
	map  { $_->[0] }
	sort { $a->[1] <=> $b->[1] }
	map  { [$_, Strassen::Util::strecke_s($_, $xy)] }
	@res;
    }
}

# incrementally try nearest with Grids => 1 to Grids => 5
# 5 can be replaced by MaxGrids parameter
sub nearest_loop {
    my($self, $x, $y, %args) = @_;
    my $max_grids = delete $args{MaxGrids} || 5;
    my @res;
    for my $grids (1 .. $max_grids) {
	@res = $self->nearest($x, $y, %args, Grids => $grids);
	return @res if (@res);
    }
    ();
}

# wie nearest, nur wird hier "x,y" als ein Argument bergeben
# XXX nearest_loop_coord fehlt
### AutoLoad Sub
sub nearest_coord {
    my($self, $xy, %args) = @_;
    my($x, $y) = split(/,/, $xy);
    $self->nearest($x, $y, %args);
}

# Zeichnet die Kreuzungen, z.B. zum Debuggen.
### AutoLoad Sub
sub draw {
    my($self, $canvas, $transpose_sub) = @_;
    $canvas->delete("crossings");
    while(my($crossing,$info) = each %{ $self->{Hash} }) {
	my($x,$y) = $transpose_sub->(split /,/, $crossing);
	$canvas->createLine($x, $y, $x, $y,
			    -tags => 'crossings',
			    -fill => 'DeepPink',
			    -capstyle => "round", # XXX see exceed bug in main
			    -width => 4,
			   );
    }
}

1;
